import os
import re
from functools import wraps
from pathlib import Path
from .helpers import HTTPMethodError
from .views import AssetView, HomeView

class URLMap:

    def __init__(self):
        self.routes = dict()
        self.secure_routes = []

    def update(self, other):
        if isinstance(other, dict):
            self.routes.update(other)
        elif isinstance(other, URLMap):
            self.secure_routes += other.secure_routes
            self.routes.update(other.routes)

    def add(self, spec, resource):
        if isinstance(spec, PathSpec):
            new_spec = spec
        else:
            new_spec = PathSpec(spec)
        return self.routes.update({new_spec : resource})

    def route_for_path(self, path):
        cands = [(k,v) for k,v in self.routes.items() if k.matches(path)]
        if not cands:
            return None, None
        else:
            cur_ps, cur_res = cands[0]
        if len(cands) == 1:
            return cur_ps, cur_res
        for pspec, resource in cands[1:]:
            if pspec > cur_ps:
                cur_ps, cur_res = pspec, resource
        return cur_ps, cur_res

    def assemble_asset_routes(self, assetDir):
        for root, dirs, names in os.walk(assetDir):
            for name in names:
                abs_path = os.path.join(root, name)
                prefix = str(Path(Path(__file__).parent.resolve()))
                path = abs_path.replace(prefix, '')
                self.add(path, Resource(AssetView(abs_path),
                                        ["GET"]))

class Resource:

    def __init__(self,
                 view_fn=None,
                 methods=["GET"],
                 roles=[]
                 ):
        self.view_fn = view_fn
        self.methods = methods
        self.roles = roles

class PathSpec:

    def __init__(self, path):
        self.spec = path

        self.is_regex = False
        self.has_groups = False
        if isinstance(self.spec, re.Pattern):
            self.is_regex = True
            if self.spec.groups >= 1:
                self.has_groups = True
        elif isinstance(self.spec, str):
            pass
        else:
            raise TypeError('Pathspec.spec must be string or re.Pattern')

    def __hash__(self):
        return hash(self.spec)

    def __str__(self):
        if self.is_regex:
            return self.spec.pattern
        return self.spec

    def __eq__(self, other):
        return str(self.spec) == str(other)

    def __gt__(self, other):
        """predescence rules go from specific to generic:
        str > capture groups > regex"""
        return (
            not self.is_regex and not \
            (not other.is_regex and not other.has_groups)
        ) or (
            self.has_groups and \
            (other.is_regex and not other.has_groups)
        )

    def __add__(self, other):

        if self.is_regex:
            spec = re.compile(str(self), str(other))
        else:
            spec = self.spec + str(other)
        return PathSpec(spec)

    def __radd__(self, other):

        if self.is_regex:
            spec = re.compile(str(other), str(self))
        else:
            spec = str(other) + self.spec
        return PathSpec(spec)

    def matches(self, path):
        if self.is_regex:
            return re.match(self.spec, path)
        else:
            return self.spec == path

    def groups(self, path):
        """retrieves groups from given path matching self.spec"""
        if self.is_regex and self.has_groups:
            return re.match(self.spec, path).groupdict()




import sys
import os
import io
import logging
import datetime
from traceback import print_tb
from io import BufferedReader
from http import HTTPStatus
from urllib.parse import (
    SplitResult,
    urlunsplit,
    quote_plus,
)
from .request import Request
from .response import Redirect
from .headers import (
    Headers,
    get_file_headers,
    get_text_headers,
    send_http_status,
)
from .session import SessionService
from .urlmap import URLMap
from .helpers import (
    strip_query_of_next_page,
    get_next_page,
)

class WSGIBase:
    """WSGI compliant application class"""

    def __init__(self, err_log=None):

        self.err_log = err_log
        self.environ = None
        self.sitename = 'example.com'
        self.session_service = SessionService(self)
        self.urlmap = URLMap()


    def __call__(self, environ, start_response):
        """
        Arguments:
            wsgi Environment: dict()
            start_response: Callable with Arguments:
                status: str()
                response_headers: list(tuple(Name: str(), Value: str()))
        returns:
            response: bytes
        """

        try:
            self.environ = environ
            request = Request(self.environ)

            # TODO: refactor request passing -> app_context usage
            self.app_context.request = request

            # creates session, either new or from token
            self.session_service.check_authentication(request)

            #____everything from here on can assume a session____

            pspec, resource = self.urlmap.route_for_path(request.path)
            if early_result := self.pre_dispatch_checks(request,
                                                        pspec,
                                                        resource):
                status, headers, response = early_result
            else:
                # hit route
                status, headers, response = self.dispatch_request(request,
                                                                  pspec,
                                                                  resource)
            # login attaches the path of the requested restricted resource,
            # or 'next=/' to query, therefore we need to catch the redirect early
            if self.needs_after_login_redirect(request):
                status, headers, response = self.after_login_redirect(request)

            # attach session cookie
            self.attach_session_cookie(headers)

            # pop request
            self.app_context.request = None

            start_response(status, headers.to_wsgi_list())
        except Exception as e:
            status, headers, response, exc_info = self.handle_exception(e)
            start_response(status, headers, exc_info)

        return response

    def handle_exception(self, e):

        s, h, r = send_http_status(HTTPStatus.INTERNAL_SERVER_ERROR)
        exc, val, tr = sys.exc_info()
        self.err_log.exception(f"{exc}, {e}")
        self.err_log.exception(f"{print_tb(tr)}")

        return s, h, [str(s).encode('utf-8')], str(exc).encode('utf-8')

    def pre_dispatch_checks(self, request, pspec, resource):
        # 404
        # TODO: Custom Error pages
        if pspec is None:
            return send_http_status(HTTPStatus.NOT_FOUND)
        # 405
        if not request.method in resource.methods:
            return send_http_status(HTTPStatus.METHOD_NOT_ALLOWED)


        # authorisation
        if self.needs_authorisation(request, resource):
            # no user logged in
            if self.session_service.user is None:
                return self.redirect_to_login(request)
            # insufficient access rights
            if not self.session_service.check_authorisation(resource):
                return send_http_status(HTTPStatus.FORBIDDEN,
                                        'Authorisation failed: No access rights.')

        # csrf protection
        if request.method in ['POST', 'PUT', 'DELETE']:
            if not self.session_service.check_csrf(request):
                return send_http_status(HTTPStatus.FORBIDDEN,
                                        'CSRF failed: Token missing or incorrect.')

    def dispatch_request(self, request, pspec, resource):

        if groups := pspec.groups(request.path):
            group_values = [v for v in groups.values()]
            response = resource.view_fn(request, *group_values)
        else:
            response = resource.view_fn(request)

        if isinstance(response, Redirect):
            return self.redirect_to_route(response.request,
                                          response.route,
                                          response.status_code)

        return self.finalize_response(response, request)

    def finalize_response(self, response, request):

        if isinstance(response, tuple) and isinstance(response[0], int):
            status = f'{response[0]}'
            response = response[1]
        else:
            s = HTTPStatus.OK
            status = f'{s.value} {s.phrase}'

        # set header based on response type
        if isinstance(response, str):
            encoded = response.encode('utf-8')
            headers = get_text_headers(encoded)

            return status, headers, [encoded]

        elif isinstance(response, BufferedReader):
            headers = get_file_headers(request, response)

            return status, headers, response

        else:
            try:
                encoded = str(response).encode('utf-8')
                headers = get_text_headers(encoded)

                return status, headers, [encoded]
            except Exception:
                raise

    def needs_authorisation(self, request, resource):
        if resource.roles and request.path != self.session_service.login_route:
            return True
        return False

    def redirect_to_login(self, request):

        request.query = (f"next={request.path}",)

        return self.redirect_to_route(request, self.session_service.login_route)

    def redirect_to_route(self, request, route, status_code=307):

        u = SplitResult(scheme=request.scheme,
                        netloc=request.host,
                        path=f"{route}",
                        query='&'.join(request.query),
                        fragment="")
        newpath = urlunsplit(u)

        headers = Headers()
        headers.add("Location", f"{newpath}")
        headers.add("Content-Length", '0')
        headers.add("Cache-Control", "no-store")

        # 307 reuses method and body of request
        # TODO: return discards body. refactor to ensure compliance
        if status_code == 307:
            s = HTTPStatus.TEMPORARY_REDIRECT
        # 303 indicates change of resource and always changes method to GET
        elif status_code == 303:
            s = HTTPStatus.SEE_OTHER

        status = f"{s.value} {s.phrase}"

        return status, headers, b''

    def attach_session_cookie(self, headers):
        """final preparation for next round trip. after headers are
        set, user and session must be invalidated."""

        if self.session_service.session: # not set for 404
            headers.add('Set-Cookie', self.session_service.session_cookie())

        self.session_service.reset_session()
        self.session_service.reset_user()

    def needs_after_login_redirect(self, request):

        # user present and '?next=' query set

        if self.session_service.user and get_next_page(request.query):
            return True
        return False

    def after_login_redirect(self, request):

        next_page = get_next_page(request.query)
        request.query = strip_query_of_next_page(request.query)

        return self.redirect_to_route(request, next_page, status_code=303)



from .urlmap import URLMap, PathSpec, Resource

class Recipe:

    def __init__(self, base_path=""):

        self.base_path = base_path
        self.urlmap = URLMap()

    def route(self, spec, methods=["GET"]):
        """
        use this decorator to register function to url_map
        under path. methods can be "POST" or "GET".
        returns Resource for further consumption by decorators.
        To be copied to app by kessel.add_recipe.
        """
        def wrapper(view_fn):
            pspec = PathSpec(spec)
            resource = Resource(view_fn, methods)
            self.urlmap.add(self.base_path + pspec, resource)
            return view_fn
        return wrapper

    def secured(self, view_fn=None, roles=['user']):
        """
        use this decorator to register ressource as restricted and assign
        roles. To be copied to app by kessel.add_recipe. Omit braces when
        using without arguments, e.g. just '@app.secure'.
        """
        def wrapper(view_fn):
            for pspec, resource in self.urlmap.routes.items():
                if view_fn is resource.view_fn:
                   resource.roles = roles
            return view_fn
        if view_fn:
            return wrapper(view_fn)
        return wrapper



import json
from uuid import uuid4 as uuid
from secrets import token_urlsafe
from datetime import timedelta, timezone
from datetime import datetime

import jwt
from jwt import ExpiredSignatureError

class SessionService:
    """Performs common operations on session objects"""

    def __init__(self, app):

        self.app = app
        self.session = None
        self.user = None
        self.login_route = '/login'
        self.cookie_key = 'session_token'
        # self.psk = token_urlsafe(512)
        # TODO: Config
        self.psk = 'change to something safe'
        # TODO: Config
        # self.private_key = self.read_key('test-app/private.pem')
        # self.private_key = self.read_key('test-app/private.pem')
        self.public_key = None
        self.public_key = None

    def read_key(self, path):
        with open(path, 'rb') as openFile:
            key = openFile.read()
        return key

    def check_authentication(self, request):
        """sets user session or initiates new one"""
        if token_msg := self.has_session_cookie(request.cookies):
            # maintain session even without user
            self.register_session(
                self.create_session(from_msg=token_msg)
            )
            # if user present, set self.user
            if uid := token_msg.get('sub', False):
                self.register_user(
                    User(uid=uid,
                         roles=token_msg.get('roles', ['user']),
                         logged_in=token_msg.get('logged_in', False))
                )
        else: # create new session
            self.register_session(self.create_session())

    def has_session_cookie(self, cookies):
        if self.cookie_key in [c.lower() for c in cookies.keys()]:
            if token_msg := self.verify_token(cookies[self.cookie_key]):
                return token_msg

    def verify_token(self, token):
        if self.psk:
            key = self.psk
        elif self.public_key:
            key = self.public_key
        try:
            msg = jwt.decode(token, key, algorithms=["HS256", "RS256"])
        except ExpiredSignatureError as e:
            # Signature expired
            # TODO: do somthing sensible here
             return None
        return msg

    def issue_token(self, msg):
        if self.psk:
            key = self.psk
            alg ="HS256"
        elif self.private_key:
            key = self.private_key
            alg = "RS256"
        args = self.jwt_args()
        return jwt.encode(msg, key, algorithm=alg)

    def create_session(self, from_msg=False):

        if msg := from_msg:
            return Session(uuid = msg['jti'],
                           csrf_token = msg['csrf_token'],
                           issued_at = msg['iat'],
                           expires = msg['exp'],
                           msg = msg)
        return Session()

    def check_authorisation(self, resource):
        """assumes user present and authenticated session"""

        # true iff any of user roles matches those tied to resource
        return any(
            map(lambda role: role in self.user.roles, resource.roles)
        )

    def check_csrf(self, request):
        # check for csrf token in header
        if csrf_token := request.headers.get('x-csrf-token'):
            return csrf_token == self.session.csrf_token

        # alternatively check in hidden form value
        if csrf_token := request.form.get('CSRFToken'):
            return csrf_token == self.session.csrf_token

    def session_cookie(self):

        if self.session.msg:
            msg = self.session.msg
        else:
            msg = {'iss' : f'{self.app.sitename}',
                   'iat' : self.session.issued_at,
                   'exp' : self.session.expires,
                   'jti' : self.session.uuid,
                   'csrf_token' : self.session.csrf_token}
        if self.user:
            msg['sub'] = self.user.uid
            msg['roles'] = self.user.roles
            msg['logged_in'] = self.user.logged_in

        cookie = (
            f"{self.cookie_key}={self.issue_token(msg)}"
            ";Path=/"
            ";HTTPOnly"
            ";SameSite=Strict"
        )
        return cookie

    def register_user(self, user):

        self.user = user
        self.app.app_context.user = user
        return True

    def reset_user(self):

        self.user = None
        self.app.app_context.user = None
        return True

    def register_session(self, session):

        self.session = session
        self.app.app_context.session = session
        return True

    def reset_session(self):

        self.session = Session()
        self.app.app_context.session = self.session
        return True

    def mock_login(self, username, password):

        self.register_user(User(uid=username, roles=['user'], logged_in=True))
        return self.user

    def jwt_args(self):
        # psk takes precedence over RSA because of sheer speed
        if self.psk:
            args = (self.psk, ["HS256"])
        elif self.public_key and self.private_key:
            args = self.public_key, ["RS256"]
        else:
            raise AttributeError('No key present')
        return args



class User:

    def __init__(self, uid=None, roles=['user'], logged_in=False):
        self.uid = uid
        self.roles = roles
        self.logged_in = logged_in

    def __str__(self):

        return str(self.uid)

class Session:
    """Object holding session information"""

    def __init__(self,
                 uuid=str(uuid()),
                 issued_at=datetime.now(timezone.utc),
                 expires=datetime.now(timezone.utc) + timedelta(hours=1),
                 csrf_token=token_urlsafe(),
                 msg=None,
                 ):

        self.uuid = uuid
        self.issued_at = issued_at
        self.expires = expires
        self.csrf_token = csrf_token
        self.msg = msg

def parse_cookies(cookie_list):
    if cookie_list:
        return dict(((c.split("=")) for c in cookie_list.split(";")))
    return dict()




from pathlib import Path
from .wsgibase import WSGIBase
from .urlmap import Resource
from .views import LoginView, LogoutView
from .context import app_context

class Kessel(WSGIBase):

    """WSGI-compliant Application Class"""

    def __init__(self, err_log=None):

        if err_log is not None:
            super().__init__(err_log=err_log)
        else:
            super().__init__(err_log=logging.getLogger('kessel.error'))

        self.urlmap.add("/login",
                         Resource(LoginView(self.session_service),
                                  ["GET", "POST"]))
        self.urlmap.add("/logout",
                         Resource(LogoutView(self.session_service),
                                  ["GET", "POST"]))
        # TODO: config
        asset_dir = Path(
            Path(__file__).parent.resolve(),
            'static'
        )
        self.urlmap.assemble_asset_routes(asset_dir)

        self.push_to_context()

    def push_to_context(self):

        app_context.app = self
        self.app_context = app_context

    def add_recipe(self, recipe):

        self.urlmap.update(recipe.urlmap)

    def route(self, spec, methods=["GET"]):
        """
        use this decorator to register function to url_map
        under spec. methods can be "POST" or "GET".
        returns Resource for further consumption by decorators,
        which is not to be called directly; kessel.dispatch_request()
        performs lookup based on route in kessel.url_map instead.
        """
        def wrapper(view_fn):
            resource = Resource(view_fn, methods)
            self.urlmap.add(spec, resource)
            return view_fn
        return wrapper

    def secured(self, view_fn=None, roles=['user']):
        """
        use this decorator to register as restricted and assign roles.
        looks up resource in url_map, adds it to url_map.secure_routes.
        returns Resource as in kessel.route. Omit braces when using
        without arguments, e.g. just '@app.secure'.
        """
        def wrapper(view_fn):
            for pspec, resource in self.urlmap.routes.items():
                if view_fn is resource.view_fn:
                   resource.roles = roles
            return view_fn
        if view_fn:
            return wrapper(view_fn)
        return wrapper





import sys
import logging
import datetime
from gunicorn import glogging


class GunicornLogger(glogging.Logger):
    """Custom logger for Gunicorn log messages."""

    # override gunicorn's now() for asctime's format
    def now(self):
            return datetime.datetime.utcnow().isoformat(sep=' ', timespec='milliseconds')

    def setup(self, cfg):
        """Configure Gunicorn application logging configuration."""

        super().setup(cfg)
        self.error_fmt =(
            '{"ts": "%(asctime)s", "loglevel": "%(levelname)s", "message": "%(message)s"}'
        )
        self.access_fmt = (
            '{'
            '"ts": "%(t)s", "host": "%(h)s", "status_line": "%(r)s", "status": "%(s)s"'
            ', "response_len": "%(b)s", "referer": "%(f)s"'
            '}'
        )

        # override gunicorn's 'access_log_format' configuration
        self.cfg.set("access_log_format", self.access_fmt)

        # Override Gunicorn's `error_log` configuration.
        self._set_handler(
            self.error_log, cfg.errorlog, logging.Formatter(
                fmt=self.error_fmt))

class MockGunicornConfig:

    def __init__(self):
        self.loglevel = 'INFO'
        self.capture_output = True
        self.errorlog = '-'
        self.accesslog = None
        self.access_log_format = ''
        self.syslog = None
        self.logconfig_dict = None
        self.logconfig_json = None
        self.logconfig = None

    def set(self, name, value):
        pass



from threading import local


class ApplicationContext(local):

    def __init__(self, /, **kw):
        self.__dict__.update(kw)
        self.app = None
        self.request = None
        self.user = None

    def current_user(self):
        return self.user

    def current_app(self):
        return self.app

    def current_request(self):
        return self.request

app_context = ApplicationContext()




from .kessel import Kessel
from .response import Redirect
from .helpers import setup_jinja2_environment
from .context import app_context

current_app = app_context.current_app
current_user = app_context.current_user
current_request = app_context.current_request




__all__ = [
    "Kessel",
    "Redirect",
    "setup_jinja2_environment",
    "current_app",
]



import os
import shutil
import mimetypes
import email
import urllib
import html
from functools import wraps
from jinja2 import Environment, FileSystemLoader

class HTTPMethodError(Exception):
    pass

def setup_jinja2_environment(templates_path='templates'):

    root = os.path.dirname(os.path.abspath(__file__))
    templates_dir = os.path.join(root, templates_path)
    env = Environment( loader = FileSystemLoader(templates_dir) )

    def render_template(template_name, **kwargs):

        template = env.get_template(template_name)
        return template.render(**kwargs)
    return render_template

def urlencoded_safe_string(urlencoded):
    return html.escape(urllib.parse.unquote_plus(urlencoded))

def get_urlencoded_data(form_string):
    out = dict()
    for kv_pairs in form_string.split('&'):
        try:
            k,v = kv_pairs.split('=')
        except ValueError:
            k = kv_pairs.split('=')
            v = ''
        key = urlencoded_safe_string(k)
        value = urlencoded_safe_string(v)
        out[key] = value
    return out

def make_multipart_message(header, form_string):
    message_string = (
        'MIME-Version: 1.0\n'
        f'Content-Type: {header}\n'
        f'{form_string}'
    )
    msg = email.message_from_string(message_string)
    if msg.is_multipart():
        return msg

def get_multipart_data(message):
    out = dict()
    for part in message.get_payload():
        name = part.get_param('name',
                              header='content-disposition')
        filename = part.get_param('filename',
                                  header='content-disposition')
        data = part.get_payload()
        if filename:
            out[html.escape(name)] = (html.escape(filename), data)
        else:
            out[html.escape(name)] = html.escape(data)
    return out

def decode_value(value):

    if isinstance(value, bytes):
        value = value.decode('latin-1')
    return str(value)

def dump_options_header(value, kwarg):

    segments = [value]
    for k,v in kwarg.items():
        if v is None:
            segments.append(k)
        else:
            # decode, then quote value
            v = decode_value(v) \
                    .replace("\\", "\\\\") \
                    .replace('"', '\\"')
            segments.append(f'{k}={v}')
    return '; '.join(segments)

def to_header_string(envstring):

    return envstring.replace('_','-').lower()

def date_time_string(timestamp=None):
    """Return the current date and time
    formatted for a message header."""
    if timestamp is None:
        timestamp = time.time()
    return email.utils.formatdate(timestamp, usegmt=True)

def get_host_from_environ(environ):

    return environ.get("HTTP_HOST") or \
            f'{environ.get("SERVER_NAME")}:{environ.get("SERVER_PORT")}'

def guess_type(path):
    """Guess the type of a file.
    Argument is a PATH (a filename).
    Return value is a string of the form type/subtype,
    usable for a MIME Content-type header.
    """
    guess, _ = mimetypes.guess_type(path)
    if guess:
        return guess
    return 'application/octet-stream'

def copy_file(source, outputfile):
    """Copy all data between two file objects.
    The SOURCE argument is a file object open for reading
    (or anything with a read() method) and the DESTINATION
    argument is a file object open for writing (or
    anything with a write() method).
    The only reason for overriding this would be to change
    the block size or perhaps to replace newlines by CRLF
    -- note however that this the default server uses this
    to copy binary data as well.
    """
    shutil.copyfileobj(source, outputfile)

def get_next_page(query_tuple):

    for query in query_tuple:
        if query.startswith("next="):
            return f"{query.replace('next=','')}"

    return None

def strip_query_of_next_page(query_tuple):

    query_list = list(query_tuple)
    for query in query_list:
        if query.startswith("next="):
            query_list.remove(query)
            break
    return tuple(query_list)

def doublewrap(f):
    '''
    a decorator decorator, allowing the decorator to be used as:
    @decorator(with, arguments, and=kwargs)
    or
    @decorator
    '''
    @wraps(f)
    def new_dec(*args, **kwargs):
        if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
            # actual decorated function
            return f(args[0])
        else:
            # decorator arguments
            return lambda realf: f(realf, *args, **kwargs)

    return new_dec



from kessel.helpers import setup_jinja2_environment


class BaseView:

    def __call__(self, request):

        return self.dispatch_request(request)

    def dispatch_request(self, request):
        pass

class AssetView(BaseView):

    def __init__(self, assetPath):

        self.path = assetPath

    def dispatch_request(self, request):

        return open(self.path, 'rb')

class HomeView(BaseView):

    def __init__(self):

        self.methods = ['GET', 'POST']

    def dispatch_request(self, request):

        return "Welcome User!"

class AuthView(BaseView):

    def __init__(self, session_service):

        self.session_service = session_service
        self.render_template = setup_jinja2_environment()

class LoginView(AuthView):

    def __init__(self, session_service):
        super().__init__(session_service)

        self.methods = ['GET', 'POST']

    def dispatch_request(self, request):

        if request.method == 'POST':

            username = request.form.get('username', 'None')
            password = request.form.get('password', 'None')

            # user = self.session_service.login_with_ldap(username, password)
            user = self.session_service.mock_login(username, password)

            if user is not None:

                return 'Login succeeded.'

        elif request.method == 'GET':

            if request.query[0]:
                action=f"{request.path}?{'&'.join(request.query)}"
            else:
                action=f"{request.path}?next=/"

            return self.render_template('login.html',
                                   action=action,
                                   csrf_token=self.session_service.session.csrf_token)
        return 'Login failed.'

class LogoutView(AuthView):

    def __init__(self, session_service):
        super().__init__(session_service)

        self.methods = ['GET', 'POST']

    def dispatch_request(self, request):

        if request.method == 'POST':
            pass
        elif request.method == 'GET':

            if self.session_service.reset_user() and self.session_service.reset_session():
                return self.render_template('logout.html')

        return 'Logout failed.'



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

requirements = [
    "cffi==1.15.1",
    "cryptography==41.0.3",
    "gunicorn==21.2.0",
    "Jinja2==3.1.2",
    "MarkupSafe==2.1.3",
    "packaging==23.1",
    "pycparser==2.21",
    "PyJWT==2.8.0",
]

setuptools.setup(
    name="kessel",
    version="0.5.2",
    author="mkirc",
    author_email="m.p.kirchner@gmx.de",
    description="a minimal wsgi framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    project_urls={
        "Source Files": "https://github.com/mkirc/kessel"
    },
    install_requires=requirements,
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3',
)




import os
import email
import datetime

from kessel.helpers import (
    decode_value,
    dump_options_header,
    to_header_string,
    guess_type,
    date_time_string,
)

def make_headers_from_environ(environ):

    h = Headers()
    h.from_environ(environ)

    return h

def send_http_status(httpstatus, msg=''):

    status = f'{httpstatus.value} {httpstatus.phrase}'
    headers = Headers()
    headers.add("Content-Type", "text/plain; charset=utf=8")

    if msg and isinstance(msg, str):
        response = msg
    else:
        response = httpstatus.description

    return status, headers, [response.encode('utf-8')]

def get_file_headers(request, response):
    """
    This sends the response code and MIME headers.
    Return value is either a file object (which has to be copied
    to the outputfile by the caller and must be closed by the
    caller under all circumstances), or None, in which case the
    caller has nothing further to do.
    """
    ctype = guess_type(request.path)
    try:
        fs = os.fstat(response.fileno())
        # Use browser cache if possible
        if ("If-Modified-Since" in request.headers \
                and "If-None-Match" not in request.headers):
            # compare If-Modified-Since and time of last file modification
            try:
                ims = email.utils.parsedate_to_datetime(
                    request.headers["If-Modified-Since"])
            except (TypeError, IndexError, OverflowError, ValueError):
                # ignore ill-formed values
                pass
            else:
                if ims.tzinfo is None:
                    # obsolete format with no timezone, cf.
                    # https://tools.ietf.org/html/rfc7231#section-7.1.1.1
                    ims = ims.replace(tzinfo=datetime.timezone.utc)
                if ims.tzinfo is datetime.timezone.utc:
                    # compare to UTC datetime of last modification
                    last_modif = datetime.datetime.fromtimestamp(
                        fs.st_mtime, datetime.timezone.utc)
                    # remove microseconds, like in If-Modified-Since
                    last_modif = last_modif.replace(microsecond=0)

                    if last_modif <= ims:
                        response.close()
                        return send_http_status(HTTPStatus.NOT_MODIFIED)
        headers = Headers()
        headers.add("Content-type", ctype)
        headers.add("Content-Length", str(fs[6]))
        headers.add("Last-Modified", date_time_string(fs.st_mtime))

        return headers

    except Exception:
        raise

def get_text_headers(response_bytes):

    headers = Headers()
    headers.add("Content-Type", "text/html", charset='utf-8')
    headers.add("Content-length", str(len(response_bytes)))

    return headers

class Headers:
    """An object that stores some headers. It has a dict-like interface,
    but is ordered, can store the same key multiple times, and iterating
    yields ``(key, value)`` pairs instead of only keys.
    This data structure is useful if you want a nicer way to handle WSGI
    headers which are stored as tuples in a list.

    Patched version of werkzeug's Headers Class.
    """

    def __init__(self):
        self._list = []

    def __getitem__(self, key):

        if isinstance(key, int):
            return self._list[key]
        elif isinstance(key, slice):
            return self.__class__(self._list[key])
        if not isinstance(key, str):
            raise KeyError()
        ikey = key.lower()
        for k, v in self._list:
            if k.lower() == ikey:
                return v


    def get(self, key, type=None):
        """Return the default value if the requested data doesn't exist.
        If `type` is provided and is a callable it should convert the value,
        return it or raise a `ValueError` if that is not possible.
        """
        try:
            rv = self.__getitem__(key)
        except KeyError:
            raise
        if type is None:
            return rv
        try:
            return type(rv)
        except ValueError:
            raise

    def getlist(self, key, type=None):
        """Return the list of items for a given key. If that key is not in the
        `Headers`, the return value will be an empty list.  Just like
        `get`, `getlist` accepts a `type` parameter.  All items will
        be converted with the callable defined there.
        """
        ikey = key.lower()
        result = []
        for k, v in self:
            if k.lower() == ikey:
                if type is not None:
                    try:
                        v = type(v)
                    except ValueError:
                        continue
                result.append(v)
        return result

    def get_all(self, name):
        """Return a list of all the values for the named field.
        This method is compatible with the `wsgiref`
        `~wsgiref.headers.Headers.get_all` method.
        """
        return self.getlist(name)

    def items(self, lower=False):
        for key, value in self:
            if lower:
                key = key.lower()
            yield key, value

    def keys(self, lower=False):
        for key, _ in self.items(lower):
            yield key

    def values(self):
        for _, value in self.items():
            yield value

    def __delitem__(self, key, _index_operation=True):
        if _index_operation and isinstance(key, (int, slice)):
            del self._list[key]
            return
        key = key.lower()
        new = []
        for k, v in self._list:
            if k.lower() != key:
                new.append((k, v))
        self._list[:] = new

    def remove(self, key):

        return self.__delitem__(key, _index_operation=False)

    def pop(self, key=None):
        """Removes and returns a key or index."""
        if key is None:
            return self._list.pop()
        if isinstance(key, int):
            return self._list.pop(key)
        try:
            rv = self[key]
            self.remove(key)
        except KeyError:
            raise
        return rv

    def __contains__(self, key):
        """Check if a key is present."""
        try:
            self.__getitem__(key)
        except KeyError:
            return False
        return True

    def __iter__(self):
        """Yield ``(key, value)`` tuples."""
        return iter(self._list)

    def __len__(self):
        return len(self._list)

    def add(self, _key, _value, **kw):
        """Add a new header tuple to the list.
        Keyword arguments can specify additional parameters for the header
        value, with underscores converted to dashes::
        """
        if kw:
            _value = dump_options_header(_value, kw)
        _key = to_header_string(_key)
        _value = decode_value(_value)
        self._validate_value(_value)
        self._list.append((_key, _value))

    def _validate_value(self, value):
        if not isinstance(value, str):
            raise TypeError("Value should be a string.")
        if "\n" in value or "\r" in value:
            raise ValueError(
                "Detected newline in header value.  This is "
                "a potential security problem"
            )

    def from_environ(self, environ):

        for k,v in environ.items():
            if k.startswith("HTTP_") and k not in (
                    "HTTP_CONTENT_TYPE",
                    "HTTP_CONTENT_LENGTH"
                    ):
                self.add(k[5:], v)
            elif k in ("CONTENT_TYPE", "CONTENT_LENGTH") and v:
                self.add(k, v)

    def add_header(self, _key, _value, **_kw):
        """Add a new header tuple to the list.
        An alias for :meth:`add` for compatibility with the :mod:`wsgiref`
        :meth:`~wsgiref.headers.Headers.add_header` method.
        """
        self.add(_key, _value, **_kw)

    def clear(self):
        """Clears all headers."""
        del self._list[:]

    def to_wsgi_list(self):
        """Convert the headers into a list suitable for WSGI.
        :return: list
        """
        return list(self)

    def __copy__(self):
        return self.__class__(self._list)

    def __str__(self):
        """Returns formatted headers suitable for HTTP transmission."""
        strs = []
        for key, value in self.to_wsgi_list():
            strs.append(f"{key}: {value}")
        strs.append("\r\n")
        return "\r\n".join(strs)

    def __repr__(self):
        return f"{type(self).__name__}({list(self)!r})"




import email
import html
from kessel.session import parse_cookies
from kessel.headers import make_headers_from_environ
from kessel.helpers import (
    get_host_from_environ,
    urlencoded_safe_string,
    make_multipart_message,
    get_urlencoded_data,
    get_multipart_data,
)


class Request:

    def __init__(self, environ):

        self.environ = environ
        self.charset = "utf-8"
        self.method = environ.get("REQUEST_METHOD",  "GET")
        self.scheme = environ.get("wsgi.url_scheme", "http")
        self.root_path = environ.get("SCRIPT_NAME", "")
        self.path = environ.get("PATH_INFO", "")
        self.query = environ.get("QUERY_STRING", ""),
        self.headers = make_headers_from_environ(self.environ)
        self.remote_addr = environ.get("REMOTE_ADDR")
        self.host = get_host_from_environ(self.environ)
        self.data = self.get_data()
        self.form = self.extract_form_data()
        self.cookies = parse_cookies(self.headers.get("Cookie"))

    def _load_data(self):
        """since wsgi.input MUST be present following PEP3333,
        we can always assume it"""
        return self.environ.get("wsgi.input")


    def _get_content_length(self):

        c_len = self.environ.get('CONTENT_LENGTH')
        if c_len is not None:
            try:
                return max(0, int(c_len))
            except (ValueError, TypeError):
                pass
        return None

    def get_data(self):

        stream = self._load_data()
        c_len = self._get_content_length()

        # A wsgi extension that tells us if the input is terminated.  In
        # that case we return the stream unchanged as we know we can safely
        # read it until the end.
        if self.environ.get("wsgi.input_terminated"):
            return stream.read()

        elif c_len is not None:
            return stream.read(c_len)

    def extract_form_data(self):

        form = dict()
        if self.data is not None:
            c_type_header = self.headers.get("content-type")
            charset = self.headers.get("charset")
            if c_type_header is not None:
                ct = c_type_header.split('; ')
                if 'application/x-www-form-urlencoded' in ct:
                    _form_str = self.data.decode(
                        charset if charset else'utf-8')
                    data = get_urlencoded_data(_form_str)
                    form.update(data)

                elif 'multipart/form-data' in ct:
                    _form_str = self.data.decode(
                        charset if charset else'utf-8')
                    msg = make_multipart_message(c_type_header, _form_str)
                    if msg is not None:
                        data = get_multipart_data(msg)
                        form.update(data)
        return form



import logging
import ldap3
from ldap3 import Server, Connection, ALL

def authenticate_search_bind(username, password):
    """
    Custom authenticator function, actually just a wrapper
    around ldap3's usual methods for search & bind.

    Args:
        username (str): Username of the user to bind (the field specified
                        as LDAP_BIND_LOGIN_ATTR)
        password (str): User's password to bind with when we find their dn.

    Returns:
        Boolean

    """

    # hardcode the server
    server = Server('ad.lfap.ph2.uni-koeln.de', get_info=ALL)
    # open anonymous connection
    connection = Connection(server, auto_bind=True)

    # hardcode the search base
    search_base = 'cn=Users,dc=lfap,dc=ph2,dc=uni-koeln,dc=de'

    # filter: is ad user
    search_filter = (
            '(&(objectClass=user)(objectCategory=person)'
            '(!(userAccountControl:1.2.840.113556.1.4.803:=2))'
            f'(uid={username}))'
    )

    # search for the user dn
    isUser = connection.search(search_base=search_base
            , search_filter=search_filter
            , attributes='uid')

    response = dict()
    result = []
    if isUser:
        result = connection.response

    if len(result) == 0:
        logging.warning(f"Authentication was not successful for user '{username}'")

    else:
        for user in result:
            # Attempt to bind with each user we find until we can find
            # one that works.

            if 'type' not in user or user.get('type') != 'searchResEntry':
                # Issue #13 - Don't return non-entry results.
                continue

            user_connection = Connection(server,
                                         user=user['dn'],
                                         password=password
                                        )

            logging.info(
                "Directly binding a connection to a server with "
                f"user:'{user['dn']}'"
                )
            try:
                bind_return_code = user_connection.bind()
                # print(f'Bind returnes: {bind_return_code}')
                logging.info(
                    f"Authentication was successful for user '{username}'"
                    )
                response['status'] = True

                # Populate User Data
                user['attributes']['dn'] = user['dn']
                response['user_info'] = user['attributes']
                response['user_id'] = username
                response['user_dn'] = user['dn']
                break

            except ldap3.core.exceptions.LDAPInvalidCredentialsResult:
                logging.info(
                    "Authentication was not successful for "
                    f"user '{username}'"
                    )
                response['status'] = False
            except Exception as e:  # pragma: no cover
                # This should never happen, however in case ldap3 does ever
                # throw an error here, we catch it and log it
                logging.error(e)
                response['status'] = False
            user_connection.unbind()
    connection.unbind()
    return response


# def validateLdap(ldapForm):
#         logging.debug('Validating LDAPLoginForm against LDAP')
#         'Validate the username/password data against ldap directory'
#         ldapManager = current_app.ldap3_login_manager
#         username = ldapForm.username.data
#         password = ldapForm.password.data

#         result = authenticate_search_bind(username, password)

#         if result.status == AuthenticationResponseStatus.success:
#             ldapForm.user = ldapManager._save_user(
#                 result.user_dn,
#                 result.user_id,
#                 result.user_info,
#                 result.user_groups
#             )
#             return True

#         else:
#             ldapForm.user = None
#             ldapForm.username.errors.append('Invalid Username/Password.')
#             ldapForm.password.errors.append('Invalid Username/Password.')
#             return False
